@@ -17,6 +17,7 @@
 
 from __future__ import absolute_import, division, print_function
 
+from collections import defaultdict
 from enum import IntEnum
 
 import attr
@@ -28,9 +29,11 @@ from hypothesis.internal.compat import (
     hbytes,
     hrange,
     int_from_bytes,
+    int_to_bytes,
     text_type,
     unicode_safe_repr,
 )
+from hypothesis.internal.conjecture.junkdrawer import IntList
 from hypothesis.internal.conjecture.utils import calc_label_from_name
 from hypothesis.internal.escalation import mark_for_escalation
 from hypothesis.utils.conventions import UniqueIdentifier
@@ -92,9 +95,15 @@ class Example(object):
     # Index of this example inside the overall list of examples.
     index = attr.ib()
 
+    # Index of the parent of this example, or None if this is the root.
+    parent = attr.ib()
+
     start = attr.ib()
     end = attr.ib(default=None)
 
+    # Index of example in parent's children, or None if this is the root.
+    child_index = attr.ib(default=None)
+
     # An example is "trivial" if it only contains forced bytes and zero bytes.
     # All examples start out as trivial, and then get marked non-trivial when
     # we see a byte that is neither forced nor zero.
@@ -109,9 +118,12 @@ class Example(object):
     # List of child examples, represented as indices into the example list.
     children = attr.ib(default=attr.Factory(list), repr=False)
 
-    @property
-    def length(self):
-        return self.end - self.start
+    # We access length a lot, and Python is annoyingly bad at basic integer
+    # arithmetic, so it makes sense to cache this on a field for speed
+    # reasons. It also reduces allocation, though most of the integers
+    # allocated from this should be easily collected garbage and/or
+    # small enough to be interned.
+    length = attr.ib(init=False, repr=False)
 
 
 @attr.s(slots=True, frozen=True)
@@ -152,6 +164,157 @@ class Block(object):
         return self.forced or self.all_zero
 
 
+class Blocks(object):
+    """A lazily calculated list of blocks for a particular ``ConjectureResult``
+    or ``ConjectureData`` object.
+
+    Pretends to be a list containing ``Block`` objects but actually only
+    contains their endpoints right up until the point where you want to
+    access the actual block, at which point it is constructed.
+
+    This is designed to be as space efficient as possible, so will at
+    various points silently transform its representation into one
+    that is better suited for the current access pattern.
+
+    In addition, it has a number of convenience methods for accessing
+    properties of the block object at index ``i`` that should generally
+    be preferred to using the Block objects directly, as it will not
+    have to allocate the actual object."""
+
+    __slots__ = ("endpoints", "owner", "__blocks", "__count", "__sparse")
+
+    def __init__(self, owner):
+        self.owner = owner
+        self.endpoints = IntList()
+        self.__blocks = {}
+        self.__count = 0
+        self.__sparse = True
+
+    def add_endpoint(self, n):
+        """Add n to the list of endpoints."""
+        assert isinstance(self.owner, ConjectureData)
+        self.endpoints.append(n)
+
+    def transfer_ownership(self, new_owner):
+        """Used to move ``Blocks`` over to a ``ConjectureResult`` object
+        when that is read to be used and we no longer want to keep the
+        whole ``ConjectureData`` around."""
+        assert isinstance(new_owner, ConjectureResult)
+        self.owner = new_owner
+        self.__check_completion()
+
+    def start(self, i):
+        """Equivalent to self[i].start."""
+        if i == 0:
+            return 0
+        else:
+            return self.end(i - 1)
+
+    def end(self, i):
+        """Equivalent to self[i].end."""
+        return self.endpoints[i]
+
+    def bounds(self, i):
+        """Equivalent to self[i].bounds."""
+        return (self.start(i), self.end(i))
+
+    def all_bounds(self):
+        """Equivalent to [(b.start, b.end) for b in self]."""
+        prev = 0
+        for e in self.endpoints:
+            yield (prev, e)
+            prev = e
+
+    def __len__(self):
+        return len(self.endpoints)
+
+    def __known_block(self, i):
+        try:
+            return self.__blocks[i]
+        except (KeyError, IndexError):
+            return None
+
+    def __getitem__(self, i):
+        n = len(self)
+        if i < -n or i >= n:
+            raise IndexError("Index %d out of range [-%d, %d)" % (i, n, n))
+        if i < 0:
+            i += n
+        assert i >= 0
+        result = self.__known_block(i)
+        if result is not None:
+            return result
+
+        # We store the blocks as a sparse dict mapping indices to the
+        # actual result, but this isn't the best representation once we
+        # stop being sparse and want to use most of the blocks. Switch
+        # over to a list at that point.
+        if self.__sparse and len(self.__blocks) * 2 >= len(self):
+            new_blocks = [None] * len(self)
+            for k, v in self.__blocks.items():
+                new_blocks[k] = v
+            self.__sparse = False
+            self.__blocks = new_blocks
+            assert self.__blocks[i] is None
+
+        start = self.start(i)
+        end = self.end(i)
+
+        # We keep track of the number of blocks that have actually been
+        # instantiated so that when every block that could be instantiated
+        # has been we know that the list is complete and can throw away
+        # some data that we no longer need.
+        self.__count += 1
+
+        # Integrity check: We can't have allocated more blocks than we have
+        # positions for blocks.
+        assert self.__count <= len(self)
+        result = Block(
+            start=start,
+            end=end,
+            index=i,
+            forced=start in self.owner.forced_indices,
+            all_zero=not any(self.owner.buffer[start:end]),
+        )
+        try:
+            self.__blocks[i] = result
+        except IndexError:
+            assert isinstance(self.__blocks, list)
+            assert len(self.__blocks) < len(self)
+            self.__blocks.extend([None] * (len(self) - len(self.__blocks)))
+            self.__blocks[i] = result
+
+        self.__check_completion()
+
+        return result
+
+    def __check_completion(self):
+        """The list of blocks is complete if we have created every ``Block``
+        object that we currently good and know that no more will be created.
+
+        If this happens then we don't need to keep the reference to the
+        owner around, and delete it so that there is no circular reference.
+        The main benefit of this is that the gc doesn't need to run to collect
+        this because normal reference counting is enough.
+        """
+        if self.__count == len(self) and isinstance(self.owner, ConjectureResult):
+            self.owner = None
+
+    def __iter__(self):
+        for i in hrange(len(self)):
+            yield self[i]
+
+    def __repr__(self):
+        parts = []
+        for i in hrange(len(self)):
+            b = self.__known_block(i)
+            if b is None:
+                parts.append("...")
+            else:
+                parts.append(repr(b))
+        return "Block([%s])" % (", ".join(parts),)
+
+
 class _Overrun(object):
     status = Status.OVERRUN
 
@@ -194,6 +357,7 @@ def calc_examples(self):
                 k = example_stack.pop()
                 ex = examples[k]
                 ex.end = index
+                ex.length = ex.end - ex.start
 
                 if ex.length == 0:
                     ex.trivial = True
@@ -210,11 +374,14 @@ def calc_examples(self):
                     label=label,
                     start=index,
                     trivial=index not in non_trivial_block_starts,
+                    parent=example_stack[-1] if example_stack else None,
                 )
                 examples.append(ex)
                 if example_stack:
                     p = example_stack[-1]
-                    examples[p].children.append(ex)
+                    children = examples[p].children
+                    ex.child_index = len(children)
+                    children.append(ex)
                 example_stack.append(i)
     for ex in examples:
         assert ex.end is not None
@@ -223,6 +390,30 @@ def calc_examples(self):
     return examples
 
 
+class DataObserver(object):
+    """Observer class for recording the behaviour of a
+    ConjectureData object, primarily used for tracking
+    the behaviour in the tree cache."""
+
+    def conclude_test(self, status, interesting_origin):
+        """Called when ``conclude_test`` is called on the
+        observed ``ConjectureData``, with the same arguments.
+
+        Note that this is called after ``freeze`` has completed.
+        """
+        pass
+
+    def draw_bits(self, n_bits, forced, value):
+        """Called when ``draw_bits`` is called on on the
+        observed ``ConjectureData``.
+        * ``n_bits`` is the number of bits drawn.
+        *  ``forced`` is True if the corresponding
+           draw was forced or ``False`` otherwise.
+        * ``value`` is the result that ``draw_bits`` returned.
+        """
+        pass
+
+
 @attr.s(slots=True)
 class ConjectureResult(object):
     """Result class storing the parts of ConjectureData that we
@@ -236,12 +427,16 @@ class ConjectureResult(object):
     example_boundaries = attr.ib()
     output = attr.ib()
     extra_information = attr.ib()
-    __examples = attr.ib(init=False, default=None)
+    has_discards = attr.ib()
+    forced_indices = attr.ib(repr=False)
+
+    __examples = attr.ib(init=False, repr=False, default=None)
 
     index = attr.ib(init=False)
 
     def __attrs_post_init__(self):
         self.index = len(self.buffer)
+        self.forced_indices = frozenset(self.forced_indices)
 
     @property
     def examples(self):
@@ -258,22 +453,35 @@ Stop = UniqueIdentifier("Stop")
 StopDiscard = UniqueIdentifier("StopDiscard")
 
 
+# Masks for masking off the first byte of an n-bit buffer.
+# The appropriate mask is stored at position n % 8.
+BYTE_MASKS = [(1 << n) - 1 for n in hrange(8)]
+BYTE_MASKS[0] = 255
+
+
 class ConjectureData(object):
     @classmethod
-    def for_buffer(self, buffer):
+    def for_buffer(self, buffer, observer=None):
         buffer = hbytes(buffer)
         return ConjectureData(
             max_length=len(buffer),
             draw_bytes=lambda data, n: hbytes(buffer[data.index : data.index + n]),
+            observer=observer,
         )
 
-    def __init__(self, max_length, draw_bytes):
+    def __init__(self, max_length, draw_bytes, observer=None):
+        if observer is None:
+            observer = DataObserver()
+        assert isinstance(observer, DataObserver)
+        self.observer = observer
         self.max_length = max_length
         self.is_find = False
         self._draw_bytes = draw_bytes
         self.overdraw = 0
-        self.block_starts = {}
-        self.blocks = []
+        self.__block_starts = defaultdict(list)
+        self.__block_starts_calculated_to = 0
+
+        self.blocks = Blocks(self)
         self.buffer = bytearray()
         self.index = 0
         self.output = u""
@@ -285,10 +493,10 @@ class ConjectureData(object):
         self.start_time = benchmark_time()
         self.events = set()
         self.forced_indices = set()
-        self.masked_indices = {}
         self.interesting_origin = None
         self.draw_times = []
         self.max_depth = 0
+        self.has_discards = False
 
         self.example_boundaries = []
 
@@ -331,7 +539,10 @@ class ConjectureData(object):
                 extra_information=self.extra_information
                 if self.extra_information.has_information()
                 else None,
+                has_discards=self.has_discards,
+                forced_indices=self.forced_indices,
             )
+            self.blocks.transfer_ownership(self.__result)
         return self.__result
 
     def __assert_not_frozen(self, name):
@@ -393,11 +604,20 @@ class ConjectureData(object):
         self.__assert_not_frozen("start_example")
         self.current_example_labels().append(label)
         self.depth += 1
-        self.max_depth = max(self.max_depth, self.depth)
+        # Logically it would make sense for this to just be
+        # ``self.depth = max(self.depth, self.max_depth)``, which is what it used to
+        # be until we ran the code under tracemalloc and found a rather significant
+        # chunk of allocation was happening here. This was presumably due to varargs
+        # or the like, but we didn't investigate further given that it was easy
+        # to fix with this check.
+        if self.depth > self.max_depth:
+            self.max_depth = self.depth
 
     def stop_example(self, discard=False):
         if self.frozen:
             return
+        if discard:
+            self.has_discards = True
         self.current_example_labels().append(StopDiscard if discard else Stop)
         self.depth -= 1
         assert self.depth >= -1
@@ -432,85 +652,96 @@ class ConjectureData(object):
         self.buffer = hbytes(self.buffer)
         self.events = frozenset(self.events)
         del self._draw_bytes
+        self.observer.conclude_test(self.status, self.interesting_origin)
 
-    def draw_bits(self, n):
+    def draw_bits(self, n, forced=None):
+        """Return an ``n``-bit integer from the underlying source of
+        bytes. If ``forced`` is set to an integer will instead
+        ignore the underlying source and simulate a draw as if it had
+        returned that integer."""
         self.__assert_not_frozen("draw_bits")
         if n == 0:
-            result = 0
-        elif n % 8 == 0:
-            return int_from_bytes(self.draw_bytes(n // 8))
+            return 0
+        assert n > 0
+        n_bytes = bits_to_bytes(n)
+        self.__check_capacity(n_bytes)
+
+        if forced is not None:
+            buf = bytearray(int_to_bytes(forced, n_bytes))
         else:
-            n_bytes = (n // 8) + 1
-            self.__check_capacity(n_bytes)
             buf = bytearray(self._draw_bytes(self, n_bytes))
-            assert len(buf) == n_bytes
-            mask = (1 << (n % 8)) - 1
-            buf[0] &= mask
-            self.masked_indices[self.index] = mask
-            buf = hbytes(buf)
-            self.__write(buf)
-            result = int_from_bytes(buf)
+        assert len(buf) == n_bytes
+
+        # If we have a number of bits that is not a multiple of 8
+        # we have to mask off the high bits.
+        buf[0] &= BYTE_MASKS[n % 8]
+        buf = hbytes(buf)
+        result = int_from_bytes(buf)
+
+        self.observer.draw_bits(n, forced is not None, result)
+
+        self.start_example(DRAW_BYTES_LABEL)
+        initial = self.index
+
+        self.buffer.extend(buf)
+        self.index = len(self.buffer)
+
+        if forced is not None:
+            self.forced_indices.update(hrange(initial, self.index))
+
+        self.blocks.add_endpoint(self.index)
+
+        self.stop_example()
 
         assert bit_length(result) <= n
         return result
 
+    @property
+    def block_starts(self):
+        while self.__block_starts_calculated_to < len(self.blocks):
+            i = self.__block_starts_calculated_to
+            self.__block_starts_calculated_to += 1
+            u, v = self.blocks.bounds(i)
+            self.__block_starts[v - u].append(u)
+        return self.__block_starts
+
+    def draw_bytes(self, n):
+        """Draw n bytes from the underlying source."""
+        return int_to_bytes(self.draw_bits(8 * n), n)
+
     def write(self, string):
-        string = hbytes(string)
+        """Write ``string`` to the output buffer."""
         self.__assert_not_frozen("write")
-        self.__check_capacity(len(string))
-        original = self.index
-        self.__write(string, forced=True)
-        self.forced_indices.update(hrange(original, self.index))
-        return string
+        string = hbytes(string)
+        if not string:
+            return
+        self.draw_bits(len(string) * 8, forced=int_from_bytes(string))
+        return self.buffer[-len(string) :]
 
     def __check_capacity(self, n):
         if self.index + n > self.max_length:
-            self.overdraw = self.index + n - self.max_length
-            self.status = Status.OVERRUN
-            self.freeze()
-            raise StopTest(self.testcounter)
+            self.mark_overrun()
 
-    def __write(self, result, forced=False):
-        self.start_example(DRAW_BYTES_LABEL)
-        initial = self.index
-        n = len(result)
-
-        block = Block(
-            start=initial,
-            end=initial + n,
-            index=len(self.blocks),
-            forced=forced,
-            all_zero=not any(result),
-        )
-
-        self.block_starts.setdefault(n, []).append(block.start)
-        self.blocks.append(block)
-        assert self.blocks[block.index] is block
-        assert len(result) == n
-        assert self.index == initial
-        self.buffer.extend(result)
-        self.index = len(self.buffer)
-        self.stop_example()
-
-    def draw_bytes(self, n):
-        self.__assert_not_frozen("draw_bytes")
-        if n == 0:
-            return hbytes(b"")
-        self.__check_capacity(n)
-        result = self._draw_bytes(self, n)
-        assert len(result) == n
-        self.__write(result)
-        return hbytes(result)
+    def conclude_test(self, status, interesting_origin=None):
+        assert (interesting_origin is None) or (status == Status.INTERESTING)
+        self.__assert_not_frozen("conclude_test")
+        self.interesting_origin = interesting_origin
+        self.status = status
+        self.freeze()
+        raise StopTest(self.testcounter)
 
     def mark_interesting(self, interesting_origin=None):
-        self.__assert_not_frozen("mark_interesting")
-        self.interesting_origin = interesting_origin
-        self.status = Status.INTERESTING
-        self.freeze()
-        raise StopTest(self.testcounter)
+        self.conclude_test(Status.INTERESTING, interesting_origin)
 
     def mark_invalid(self):
-        self.__assert_not_frozen("mark_invalid")
-        self.status = Status.INVALID
-        self.freeze()
-        raise StopTest(self.testcounter)
+        self.conclude_test(Status.INVALID)
+
+    def mark_overrun(self):
+        self.conclude_test(Status.OVERRUN)
+
+
+def bits_to_bytes(n):
+    """The number of bytes required to represent an n-bit number.
+    Equivalent to (n + 7) // 8, but slightly faster. This really is
+    called enough times that that matters."""
+    return (n + 7) >> 3
